package com.treweren.parent;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Vector;
import java.util.HashSet;
import java.io.BufferedReader;
import java.io.FileReader;

import org.knime.base.node.io.filereader.FileAnalyzer;
import org.knime.base.node.io.filereader.FileReaderNodeSettings;
import org.knime.base.node.io.filereader.FileReaderSettings;
import org.knime.base.node.io.filereader.FileTable;
import org.knime.base.node.io.filereader.SettingsStatus;
import org.knime.chem.types.SdfValue;
import org.knime.chem.types.SmilesValue;
import org.knime.core.data.DataCell;
import org.knime.core.data.DataColumnSpec;
import org.knime.core.data.DataColumnSpecCreator;
import org.knime.core.data.DataRow;
import org.knime.core.data.DataTableSpec;
import org.knime.core.data.RowKey;
import org.knime.core.data.def.DefaultRow;
import org.knime.core.data.def.DoubleCell;
import org.knime.core.data.def.IntCell;
import org.knime.core.data.def.StringCell;
import org.knime.core.node.BufferedDataContainer;
import org.knime.core.node.BufferedDataTable;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.defaultnodesettings.SettingsModelBoolean;
import org.knime.core.node.defaultnodesettings.SettingsModelIntegerBounded;
import org.knime.core.node.defaultnodesettings.SettingsModelInteger;
import org.knime.core.node.defaultnodesettings.SettingsModelDoubleRange;
import org.knime.core.node.defaultnodesettings.SettingsModelString;
import org.knime.core.node.defaultnodesettings.SettingsModelDouble;
import org.knime.core.node.ExecutionContext;
import org.knime.core.node.ExecutionMonitor;
import org.knime.core.node.InvalidSettingsException;
import org.knime.core.node.NodeLogger;
import org.knime.core.node.NodeModel;
import org.knime.core.node.NodeSettingsRO;
import org.knime.core.node.NodeSettingsWO;

import org.knime.core.data.RowKey;
import org.knime.core.node.BufferedDataContainer;
import org.knime.core.util.MutableBoolean;
import org.knime.core.data.DataColumnSpecCreator;
import org.knime.chem.types.SmilesCell;
import org.knime.core.data.def.DoubleCell;
import org.knime.core.data.def.IntCell;
import org.knime.core.data.def.StringCell;
import org.knime.core.data.def.DefaultRow;
import org.knime.core.data.DataType;




/**
 * This is the model implementation of Parent.
 * Convert to parent form and apply drug-like filters
 *
 * @author Treweren Consultants
 */
public class ParentNodeModel extends NodeModel {
    
    // the logger instance
    private static final NodeLogger logger = NodeLogger
            .getLogger(ParentNodeModel.class);
        
    /** the settings key which is used to retrieve and 
        store the settings (from the dialog or from a settings file)    
       (package visibility to be usable from the dialog). */
    private static int n_Column = 0;
    private static String m_sdfColumn = null;
    private static String m_smilesColumn = null;
    private static String m_stringColumn = null;
    private static final int CANCEL_CHECK_INTERVAL = 1000;
    
 	static final String CFGKEY_FILTER = "Filter";
 	static final String CFGKEY_CENTRES = "Centres";
 	static final String CFGKEY_MASS = "Mass";
 	static final String CFGKEY_XSA = "XSA";
	static final String CFGKEY_ROTBONDS = "Rot-Bonds";
 	
	static final String CFGKEY_SMILES = "Smiles";
	static final String CFGKEY_WORKING = "Working";

	/** initial default count value. */
	static final Boolean DEFAULT_FILTER = true;
	static final int DEFAULT_CENTRE1 = 2;
	static final int DEFAULT_CENTRE2 = 9;
	static final double DEFAULT_MASS1 = 150.;
	static final double DEFAULT_MASS2 = 800.;
	static final double DEFAULT_XSA1 = 20.;
	static final double DEFAULT_XSA2 = 240.;
	static final int DEFAULT_ROTBONDS1 = 0;
	static final int DEFAULT_ROTBONDS2 = 10;
	
	static final Boolean DEFAULT_SMILES = false;
	static final String DEFAULT_WORKING =  System.getenv("THINK_WORKING");

	// example value: the models count variable filled from the dialog 
	// and used in the models execution method. The default components of the
	// dialog work with "SettingsModels".
	
	private final SettingsModelBoolean m_filter =
		new SettingsModelBoolean(ParentNodeModel.CFGKEY_FILTER,
                    ParentNodeModel.DEFAULT_FILTER );
	
	private final SettingsModelDoubleRange m_centres =
		new SettingsModelDoubleRange(ParentNodeModel.CFGKEY_CENTRES, ParentNodeModel.DEFAULT_CENTRE1, ParentNodeModel.DEFAULT_CENTRE2 );

	private final SettingsModelDoubleRange m_mass =
		new SettingsModelDoubleRange(ParentNodeModel.CFGKEY_MASS, ParentNodeModel.DEFAULT_MASS1, ParentNodeModel.DEFAULT_MASS2  );

	private final SettingsModelDoubleRange m_xsa =
		new SettingsModelDoubleRange(ParentNodeModel.CFGKEY_XSA, ParentNodeModel.DEFAULT_XSA1, ParentNodeModel.DEFAULT_XSA2 );

	private final SettingsModelDoubleRange m_rotbonds =
		new SettingsModelDoubleRange(ParentNodeModel.CFGKEY_ROTBONDS, ParentNodeModel.DEFAULT_ROTBONDS1, ParentNodeModel.DEFAULT_ROTBONDS2 );

	private final SettingsModelBoolean m_smiles =
		new SettingsModelBoolean(ParentNodeModel.CFGKEY_SMILES,
                    ParentNodeModel.DEFAULT_SMILES );

	private final SettingsModelString m_working =
		new SettingsModelString(ParentNodeModel.CFGKEY_WORKING, ParentNodeModel.DEFAULT_WORKING );


    /**
     * Constructor for the node model.
     */
    protected ParentNodeModel() {
    
        // One incoming port and one outgoing port
        super(1, 1);
    }

    /**
     * @see org.knime.core.node.NodeModel #execute(BufferedDataTable[],
     *      ExecutionContext)
     */
    @Override
    protected DataTableSpec[] configure(final DataTableSpec[] inSpecs) throws InvalidSettingsException {

        if ( System.getenv("THINK_EXEC") == null ) {
            throw new InvalidSettingsException(
            "Environment variable THINK_EXEC is unset - It should be set to the folder containing the THINK software.");       	  		
    	}
        if ( System.getenv("THINK_WORKING") == null ) {
            throw new InvalidSettingsException(
            "Environment variable THINK_WORKING is not set to the folder in which working files will be created.");       	  		
    	}

        DataTableSpec in = inSpecs[0];
        n_Column = columnSelect (in);
        return new DataTableSpec[] {null};
    }
    protected int columnSelect ( final DataTableSpec in ) throws  InvalidSettingsException {

        StringBuilder warningMessage = new StringBuilder();
        int n_column=0;
        int sdfColCount = 0;
        int n_sdf=0;
        int smilesColCount = 0;
        int n_smiles=0;
        int stringColCount = 0;
        int n_string=0;
        for (int i = 0; i < in.getNumColumns(); i++) {
            DataColumnSpec s = in.getColumnSpec(i);
            if (s.getType().isCompatible(SdfValue.class)) {
                if (m_sdfColumn == null) {
                    m_sdfColumn = in.getColumnSpec(i).getName();
                    n_sdf = i;
                }
                sdfColCount++;
            }
            else if ( s.getType().isCompatible(SmilesValue.class)) {
            	if (m_smilesColumn == null ) {
             		m_smilesColumn = in.getColumnSpec(i).getName();
             		n_smiles=i;
              	}
               	smilesColCount++;
            }
            else if ( in.getColumnSpec(i).getName().equalsIgnoreCase("smiles"))  {
            	if (m_stringColumn == null ) {
             		m_stringColumn = in.getColumnSpec(i).getName();
             		n_string=i;
              	}
               	stringColCount++;
            }
        }
        if ( smilesColCount == 0 & sdfColCount == 0 & stringColCount == 0) {
            throw new InvalidSettingsException(
                    "No column with Smiles or SD file compatible type");       	
        }
        else if ( ( m_smiles.getBooleanValue() | sdfColCount == 0 ) & smilesColCount > 0 ) {
        	m_sdfColumn = null;
        	m_stringColumn = null;
        	n_column = n_smiles;
        	if (smilesColCount > 1) {
                warningMessage.append("More than one Smiles compatible column in " 
                        + "input, using column \"" + m_smilesColumn + "\".");
        	}
        }
        else if ( sdfColCount > 0 ) {
           	m_smilesColumn = null;
        	m_stringColumn = null;
        	n_column = n_sdf;
        	if (sdfColCount > 1) {
                warningMessage.append("More than one SDF compatible column in " 
                        + "input, using column \"" + m_sdfColumn + "\".");
        	}
        }
        else if ( stringColCount > 0 ) {
        	m_smilesColumn = null;
        	m_sdfColumn = null;
        	n_column = n_string;
           	if (stringColCount > 1) {
                warningMessage.append("More than one Smiles string compatible column in " 
                        + "input, using column \"" + m_stringColumn + "\".");
           	}
        }
        if (warningMessage.length() > 0) {
            setWarningMessage(warningMessage.toString());
        }
        return n_column;
    }

 
    /**
     * @see org.knime.core.node.NodeModel #execute(BufferedDataTable[],
     *      ExecutionContext)
     */
    @Override
    protected BufferedDataTable[] execute(final BufferedDataTable[] inData,
            final ExecutionContext exec) throws Exception {

        notifyViews(null);

        final String cwdString = m_working.getStringValue();
        // clean up 
        File outputReport = new File(cwdString, "output.log"); 
        outputReport.delete();
        File outputFile = new File(cwdString, "parent.csv"); 
        outputFile.delete();
        BufferedDataTable in = inData[0];

        if ( n_Column == 0 ) {
        	DataTableSpec InSpec = in.getDataTableSpec();
        	n_Column = columnSelect(InSpec);
        }


        // write out input file
        String fileString = null; 
        if ( m_sdfColumn == null ) {
            fileString = "parent.smiles"; 
        }
        else {
            fileString = "parent.sdf"; 
        }
        File inFile = new File(cwdString, fileString); 
        BufferedWriter outWriter = new BufferedWriter(new FileWriter(inFile));
         int colIndex;
        if ( m_sdfColumn != null ) {
        	colIndex = in.getDataTableSpec().findColumnIndex(m_sdfColumn);
        }
        else if ( m_smilesColumn != null ) {
        	colIndex = in.getDataTableSpec().findColumnIndex(m_smilesColumn);        
        }
        else {
            colIndex = in.getDataTableSpec().findColumnIndex(m_stringColumn);        
        }
        	  
        final double count = in.getRowCount(); // floating point operations
        int i = 0; 
        int missingCount = 0;
        for (DataRow r : in) {
            exec.checkCanceled();
            DataCell c = r.getCell(colIndex);
            if (c.isMissing()) {
                missingCount++;
            } 
            else if ( m_sdfColumn != null ) {
                SdfValue v = (SdfValue)c;
                String toString = v.toString();
                outWriter.write(toString);
                if (!toString.trim().endsWith("$$$$")) {
                    outWriter.newLine();
                    outWriter.append("$$$$");
                    outWriter.newLine();
                }
            }
            else if ( m_smilesColumn != null ){
                SmilesValue v = (SmilesValue)c;
                String toString = v.toString();
                outWriter.write(toString);        	
                outWriter.newLine();
            }
            else {
            	String toString = c.toString() + " " + r.getKey();
            	outWriter.write(toString);        	
            	outWriter.newLine();
            }
            i++;
        }

        outWriter.close();
        if (missingCount > 0) {
            setWarningMessage("Skipped " + missingCount 
                    + " row(s) because of missing values");
        }

        // execute THINK
        File cwdFile = new File (cwdString);
        int exitVal;
        try {
            exec.setProgress(0., "Starting THINK");
            Runtime rt = Runtime.getRuntime();
            String s_filter = "NONE";
            if ( m_filter.getBooleanValue() ) {
              s_filter = "FILTER";
            }
            double n_centre1 = m_centres.getMinRange();
            double n_centre2 = m_centres.getMaxRange();
            double d_mass1 = m_mass.getMinRange();
            double d_mass2 = m_mass.getMaxRange();
            double d_xsa1 = m_xsa.getMinRange();
            double d_xsa2 = m_xsa.getMaxRange();
            double n_rotbonds1 = m_rotbonds.getMinRange();
            double n_rotbonds2 = m_rotbonds.getMaxRange();
            
            
            // prepare the command string
            String cmdString = System.getenv("THINK_EXEC") + "think THINK_EXEC:parent.log output.log " + fileString + " " + s_filter
            + " " + n_centre1 + " " + n_centre2 + " " + d_mass1 + " " + d_mass2 + " " + d_xsa1 + " " + d_xsa2 + " " + n_rotbonds1 + " " + n_rotbonds2;
            // Go!
            logger.info("THINK command line: '" + cmdString + "'");
            final Process proc = rt.exec(cmdString, null, cwdFile);

            final MutableBoolean procDone = new MutableBoolean(false);
            new Thread(new Runnable() {
                public void run() {
                    synchronized (procDone) {
                        while (!procDone.booleanValue()) {
                            try {
                                exec.checkCanceled();
                            } catch (CanceledExecutionException cee) {
                                // blow away the running external process
                                proc.destroy();
                                return;
                            }
                            try {
                               	procDone.wait(CANCEL_CHECK_INTERVAL);
                                // Read progress file and set progress value       
                               	double d= 0.;
                                File progressReport = new File(cwdString, "progress0.dat");
                                if ( progressReport.exists() && progressReport.isFile() ) {
                               		if ( progressReport.length() > 0 ) {
                               			BufferedReader in = new BufferedReader(new FileReader(progressReport));
                               			String line;
                               			if ( (line = in.readLine()) != null) {
                               			  d = Double.valueOf(line).doubleValue();
                               			}               
                              			if ( (line = in.readLine()) == null) {
                              				line = "Running THINK";
                              			}
                               			in.close();
                                        exec.setProgress( d, line.toString());
                               		}
                                }
                            } catch (InterruptedException e) {
                                // do nothing
                            } catch ( FileNotFoundException e) {
                                // do nothing
                            } catch ( IOException e) {
                                // do nothing
                            }
                        }
                    }
                }

            }).start();
           

            // wait until the external process finishes.
            exitVal = proc.waitFor();

            synchronized (procDone) {
                // this should terminate the check cancel thread
                procDone.setValue(true);
            }

            exec.checkCanceled();

            exec.setProgress("Wrapping up");
            logger.info("THINK terminated with exit code: " + exitVal);
        } catch (Throwable t) {
        	warningReport();
        	logger.error("THINK failed (with exception)", t);
            throw new Exception(t);
        }

        if (exitVal != 0) {
            // before we return barfing, we save the output in the failing list
        	warningReport();
            throw new IllegalStateException("THINK failed (error code "
                    + exitVal + ")");
        }

        exec.setProgress("Creating KNIME tables");
        BufferedDataTable outTable = readCSVFile(exec);

        return new BufferedDataTable[]{outTable};

    }

    private void warningReport( ) throws Exception {

    	StringBuilder warningMessage = new StringBuilder();
        String cwdString = m_working.getStringValue();
        File outputReport = new File(cwdString, "output.log"); 

        if ( outputReport.exists() && outputReport.isFile() ) {
       		if ( outputReport.length() > 0 ) {
       			BufferedReader in = new BufferedReader(new FileReader(outputReport));
       			String line;
       			while ((line = in.readLine()) != null) {
       				warningMessage.append( line + "\n");
       			}               
       			in.close();
       		}
       		else { 
                warningMessage.append( "Report from THINK is blank");      	     			
       		}
        }
       	else {
            warningMessage.append( "No report from THINK");      	
       	}	
        if (warningMessage.length() > 0) {
            setWarningMessage(warningMessage.toString());
        }
   }
    private BufferedDataTable readCSVFile(final ExecutionContext exec) throws Exception { 

        String cwdString =  m_working.getStringValue();
        File outFile = new File (cwdString, "parent.csv");
        if (!(outFile.exists() && outFile.isFile())) {
        	warningReport();
        	throw new IllegalStateException("THINK didn't produce any output"
                    + " at the specified location ('"
                    + outFile.getAbsolutePath() + "')");
        }
        HashSet<String> titles = new HashSet<String>();
        String title;
        String line;
        boolean headerRead = false;
        boolean columnsCreated = false;
        DataColumnSpec[] allColSpecs;
        BufferedReader in = new BufferedReader(new FileReader(outFile));
        DataTableSpec outputSpec;
        BufferedDataContainer cont =null;
        RowKey key =null;
        int count =0;
        int colcount =0;
        int rowcount =0;
        char chr='0';
        int keys = 0;
        int ilast = 0;
        String fieldNames[] = new String[100];
        int fieldTypes[] = new int[100];
        while ((line = in.readLine()) != null) {
            exec.checkCanceled();
// Extract column headers
            if ( !headerRead ) {
                headerRead = true;
                colcount =0;
                for ( int i=0; i<=line.length(); i++ ) {
                    if ( i < line.length() ) 
                        chr = line.charAt(i);
                    else
                        chr = ',';
                    if ( chr == ',' ) { 
                        fieldNames[colcount] = line.substring(ilast,i);
                        if ( line.substring(ilast,i).equalsIgnoreCase("SMILES"))
                        	fieldTypes[colcount]= 2;
                        else if ( line.substring(ilast,i).equalsIgnoreCase("NAME"))
                        	fieldTypes[colcount]= 1;
                        else if ( line.substring(ilast,i).equalsIgnoreCase("KEYS"))
                        	fieldTypes[colcount]= 3;
                        ilast = i+1;
                        colcount++;
                    }
                }
                colcount--;
            }
// Load data
            else {
                DataCell[] cells = new DataCell[colcount];
                rowcount++;
                count =0;                 
                double d_value;
                 for ( int i=0; i<=line.length(); i++ ) {
                    if ( i < line.length() ) 
                        chr = line.charAt(i);
                    else
                        chr = ',';
                    if ( chr == ',' ) { 
                        if ( count == 0 ) {
                            title = line.substring(0,i);
                            if (title.length() == 0) 
                                title = "Mol " + rowcount;
                            if (titles.contains(title)) 
                                title = title + "-" + rowcount;
                            titles.add(title);
                            key = new RowKey(title);                           
                        }
                        else if  ( fieldTypes[count] == 2 ) 
                            cells[count-1] = new SmilesCell(line.substring(ilast,i));
                        else if  ( fieldTypes[count]== 3) 
                            cells[count-1] = new StringCell(line.substring(ilast,i));
                        else if ( count <= colcount ) {
                          try {
                            d_value = Double.valueOf(line.substring(ilast,i)).doubleValue();
                            cells[count-1] = new DoubleCell(d_value);
                            if ( fieldTypes[count] == 0 )
                            	fieldTypes[count] = 4;
                          } catch (Throwable t) {
                            if ( fieldTypes[count] == 0 ) {
                            	fieldTypes[count] = 3;
                                cells[count-1] = new StringCell(line.substring(ilast,i));
                            }
                            else
                                cells[count-1] = DataType.getMissingCell();
                          }
                        }
                        count ++;
                        ilast = i + 1;
                    }
                 }
  // Create columns?
                 if (!columnsCreated )
                 {
                     allColSpecs = new DataColumnSpec[colcount];
                     columnsCreated = true;
  // count= 0 will be molecule names
                     for (count=1; count<=colcount; count++)
                     {                   	 
                       if  ( fieldTypes[count] == 2 ) 
                         allColSpecs[count-1] = new DataColumnSpecCreator(fieldNames[count],SmilesCell.TYPE).createSpec();
                       else if  ( fieldTypes[count] == 4 ) 
                         allColSpecs[count-1] = new DataColumnSpecCreator(fieldNames[count], DoubleCell.TYPE).createSpec();
                       else if ( fieldTypes[count] == 3 )
                         allColSpecs[count-1] = new DataColumnSpecCreator(fieldNames[count],StringCell.TYPE).createSpec();
                       else
                    	 allColSpecs[count-1] = new DataColumnSpecCreator("Unknown",StringCell.TYPE).createSpec();
                     }
                     outputSpec = new DataTableSpec(allColSpecs);
                     cont = exec.createDataContainer(outputSpec);               	 
                 }
                DefaultRow row = new DefaultRow(key, cells);
                cont.addRowToTable(row);
            }
        }
        cont.close();
        in.close();

        return cont.getTable();
    }

    private BufferedDataTable readOutputFile(final ExecutionContext exec) throws Exception { 

        String cwdString =  m_working.getStringValue();
        File outFile = new File (cwdString, "parent.csv");
        if (!(outFile.exists() && outFile.isFile())) {
        	warningReport();
        	throw new IllegalStateException("THINK didn't produce any output"
                    + " at the specified location ('"
                    + outFile.getAbsolutePath() + "')");
        }
        HashSet<String> titles = new HashSet<String>();
        String title;
        String line;
        boolean headerRead = false;
        DataColumnSpec[] allColSpecs;
        BufferedReader in = new BufferedReader(new FileReader(outFile));
        DataTableSpec outputSpec;
        BufferedDataContainer cont =null;
        RowKey key =null;
        int count =0;
        int rowcount =0;
        char chr='0';
        int keys = 0;
        while ((line = in.readLine()) != null) {
            exec.checkCanceled();
// Extract column headers
            if ( !headerRead ) {
                headerRead = true;
                count =0;
                for ( int i=0; i<line.length(); i++ ) {
                    if ( line.charAt(i) == ',' ) count ++;
                }
// Create column definitions
                allColSpecs = new DataColumnSpec[1];
                count =0;
                int ilast=0;
                for ( int i=0; i<=line.length(); i++ ) {
                    if ( i < line.length() ) {
                        chr = line.charAt(i);
                    }
                    else {
                        chr = ',';
                    }
                    if ( chr == ',' ) {
                        if  ( count == 1 ) {
                            allColSpecs[count-1] = 
                            new DataColumnSpecCreator(line.substring(ilast,i),SmilesCell.TYPE).createSpec();
                            break;
                        }
                        else if  ( line.substring(ilast,i).equals("KEYS") ) {
                            keys = count;
                            allColSpecs[count-1] = 
                            new DataColumnSpecCreator(line.substring(ilast,i),StringCell.TYPE).createSpec();
                        }
                        else if ( count > 1 ) {
                            allColSpecs[count-1] = 
                            new DataColumnSpecCreator(line.substring(ilast,i), DoubleCell.TYPE).createSpec();
                        }
                        count ++;
                        ilast = i + 1;
                    }
                }
                outputSpec = new DataTableSpec(allColSpecs);
                cont = exec.createDataContainer(outputSpec);
            }
// Load data
            else {
                DataCell[] cells = new DataCell[1];
                rowcount++;
                count =0;                  
                double d_value;
                int ilast=0;
                for ( int i=0; i<=line.length(); i++ ) {
                    if ( i < line.length() ) {
                        chr = line.charAt(i);
                    }
                    else
                        chr = ',';
                    if ( chr == ',' ) { 
                        if ( count == 0 ) {
                            title = line.substring(0,i);
                            if (title.length() == 0) {
                                title = "Mol " + rowcount;
                            }
                            if (titles.contains(title)) {
                                title = title + "-" + rowcount;
                            }
                            titles.add(title);
                            key = new RowKey(title);                           
                        }
                        else if  ( count == 1 ) {
                            cells[count-1] = new SmilesCell(line.substring(ilast,i));
                            break;
                        }
                        else if  ( count == keys ) {
                            cells[count-1] = new StringCell(line.substring(ilast,i));
                        }
                        else {
                        try {
                            d_value = Double.valueOf(line.substring(ilast,i)).doubleValue();
                            cells[count-1] = new DoubleCell(d_value);
                        } catch (Throwable t) {
                            cells[count-1] = DataType.getMissingCell();
                        }
                        }
                        count ++;
                        ilast = i + 1;
                    }
                }
                DefaultRow row = new DefaultRow(key, cells);
                cont.addRowToTable(row);
            }
        }
        cont.close();
        in.close();

        return cont.getTable();
    }    /**
     * @see org.knime.core.node.NodeModel#reset()
     */
    @Override
    protected void reset() {
        // TODO Code executed on reset.
        // Models build during execute are cleared here.
        // Also data handled in load/saveInternals will be erased here.
    }

    /**
     * @see org.knime.core.node.NodeModel
     *      #configure(org.knime.core.data.DataTableSpec[])
     */
 
    /**
     * @see org.knime.core.node.NodeModel
     *      #saveSettingsTo(org.knime.core.node.NodeSettings)
     */
    @Override
    protected void saveSettingsTo(final NodeSettingsWO settings) {

        // TODO save user settings to the config object.
		
		m_filter.saveSettingsTo(settings);
		m_centres.saveSettingsTo(settings);
		m_mass.saveSettingsTo(settings);
		m_xsa.saveSettingsTo(settings);
		m_rotbonds.saveSettingsTo(settings);
		m_smiles.saveSettingsTo(settings);
		m_working.saveSettingsTo(settings);

    }

    /**
     * @see org.knime.core.node.NodeModel
     *      #loadValidatedSettingsFrom(org.knime.core.node.NodeSettingsRO)
     */
    @Override
    protected void loadValidatedSettingsFrom(final NodeSettingsRO settings)
            throws InvalidSettingsException {
            
        // TODO load (valid) settings from the config object.
        // It can be safely assumed that the settings are valided by the 
        // method below.
        
        m_filter.loadSettingsFrom(settings);
        m_centres.loadSettingsFrom(settings);
        m_mass.loadSettingsFrom(settings);
        m_xsa.loadSettingsFrom(settings);
        m_rotbonds.loadSettingsFrom(settings);
        m_smiles.loadSettingsFrom(settings);
        m_working.loadSettingsFrom(settings);

    }

    /**
     * @see org.knime.core.node.NodeModel
     *      #validateSettings(org.knime.core.node.NodeSettingsRO)
     */
    @Override
    protected void validateSettings(final NodeSettingsRO settings)
            throws InvalidSettingsException {
            
        // TODO check if the settings could be applied to our model
        // e.g. if the count is in a certain range (which is ensured by the
        // SettingsModel).
        // Do not actually set any values of any member variables.

        m_filter.validateSettings(settings);
        m_centres.validateSettings(settings);
        m_mass.validateSettings(settings);
        m_xsa.validateSettings(settings);
        m_rotbonds.validateSettings(settings);
        m_smiles.validateSettings(settings);
        m_working.validateSettings(settings);

    }
    
    /**
     * @see org.knime.core.node.NodeModel #loadInternals(java.io.File,
     *      org.knime.core.node.ExecutionMonitor)
     */
    @Override
    protected void loadInternals(final File internDir,
            final ExecutionMonitor exec) throws IOException,
            CanceledExecutionException {
        
		// TODO load internal data. 
		// Everything handed to output ports is loaded automatically (data
		// returned by the execute method, models loaded in loadModelContent,
		// and user settings set through loadSettingsFrom - is all taken care 
		// of). Load here only the other internals that need to be restored
		// (e.g. data used by the views).

    }
    
    /**
     * @see org.knime.core.node.NodeModel #saveInternals(java.io.File,
     *      org.knime.core.node.ExecutionMonitor)
     */
    @Override
    protected void saveInternals(final File internDir,
            final ExecutionMonitor exec) throws IOException,
            CanceledExecutionException {
       
       	// TODO save internal models. 
		// Everything written to output ports is saved automatically (data
		// returned by the execute method, models saved in the saveModelContent,
		// and user settings saved through saveSettingsTo - is all taken care 
		// of). Save here only the other internals that need to be preserved
		// (e.g. data used by the views).

    }

}
